from datetime import timedelta


def timedelta_isoformat(td: timedelta) -> str:
    """
    ISO 8601 encoding for Python timedelta object.
    Taken from pydantic source:
        https://github.com/pydantic/pydantic/blob/3704eccce4661455acdda1cdcf716bd4b3382e08/pydantic/deprecated/json.py#L135-L140

    """
    minutes, seconds = divmod(td.seconds, 60)
    hours, minutes = divmod(minutes, 60)
    return f'{"-" if td.days < 0 else ""}P{abs(td.days)}DT{hours:d}H{minutes:d}M{seconds:d}.{td.microseconds:06d}S'



from typing import Any, List, Literal, Optional, TypedDict


class User(TypedDict):
    code: str
    username: Optional[str]
    description: str
    tenant: str
    permission_group: str


class HiddenUserAPIKey(TypedDict):
    code: str
    description: str
    last_used: Optional[str]
    custom_policy: Optional[str]


class UserDetails(TypedDict):
    code: str
    username: Optional[str]
    first_name: Optional[str]
    last_name: Optional[str]
    email_address: Optional[str]
    mobile_number: Optional[str]

    api_keys: list[HiddenUserAPIKey]


class NewUserApiKey(TypedDict):
    code: str
    new_api_key: str
    description: str
    custom_policy: Optional[str]


class Tenant(TypedDict):
    code: str
    name: str
    description: str
    logo_url: Optional[str]


class AccessControlListJsonEntry(TypedDict):
    action: str
    resource: str


class AccessControlListJson(TypedDict):
    version: Literal[1]
    permissions: list[AccessControlListJsonEntry]


class PermissionGroup(TypedDict):
    tenant: str
    code: str
    label: str
    acl_policy: str


class VerifyMeResult(TypedDict):
    user: Optional[str]
    tenant: Optional[str]
    permission_group: Optional[str]
    acl_policy: str


HealthCheckMessageResult = Literal['OK', 'ERROR']


class SystemHealthcheckResult(TypedDict):
    database: HealthCheckMessageResult
    cache: HealthCheckMessageResult
    ephemeral_document_store: HealthCheckMessageResult
    ephemeral_warehouse: HealthCheckMessageResult


class AuthHealthcheckResult(TypedDict):
    status: HealthCheckMessageResult


class ProjectStatus(TypedDict):
    last_ping: Optional[str]
    last_sample: Optional[str]
    ip_address: Optional[str]


class Project(TypedDict):
    code: str
    name: str
    project_number: str
    active: bool
    description: str
    latitude: Optional[float]
    longitude: Optional[float]
    timezone: str
    display_timezone: Optional[str]
    tenant: str
    status: ProjectStatus


class ProjectCode(TypedDict):
    code: str


class ItemCode(TypedDict):
    code: str


class Units(TypedDict):
    code: str
    name: str
    unit: Optional[str]


CHANNEL_CLASSIFICATION = Literal['PRIMARY', 'SECONDARY',
                                 'EVENT_METRIC', 'PRIMARY_EVENT_AGGREGATE']


class Channel(TypedDict):
    code: str
    project: ItemCode
    units: Optional[Units]
    sampling_period: Optional[str]
    classification: CHANNEL_CLASSIFICATION
    label: str
    dtype: str
    device: Optional[ItemCode]
    metric: Optional[str]
    source: Optional[ItemCode]
    aggregate: Optional[str]
    value_range_min: Optional[float]
    value_range_max: Optional[float]
    channel_multiplier: float
    channel_offset: float
    last_valid_timestamp: Optional[str]


class DeviceType(TypedDict):
    code: str
    description: str
    manufacturer: str
    model_number: str


class Device(TypedDict):
    code: str
    project: ProjectCode
    label: str
    location_description: Optional[str]
    device_type: DeviceType
    groups: List[str]


class DataSample(TypedDict):
    timestamp: str
    channel_code: str
    value: float


class WidgetConfig(TypedDict):
    type: str
    config: dict


class WidgetColumn(TypedDict):
    size: Optional[str | int]
    widget: WidgetConfig


class WidgetRow(TypedDict):
    columns: List[WidgetColumn]
    height: int
    title: str
    breakpoint: Optional[str]


class Dashboard(TypedDict):
    icon: Optional[str]
    name: Optional[str]
    banner_image: Optional[str]
    widgets: Optional[List[WidgetRow]]
    fullscreen: Optional[bool]


class Dashboards(TypedDict):
    dashboards: List[Dashboard]


class DataRequestMeta(TypedDict):
    first_timestamp: str
    last_timestamp: str


class DataRequest(TypedDict):
    code: ItemCode
    requested_at: str
    completed_at: str | None
    in_progress: bool
    presigned_url: str | None
    message: str | None
    mime_type: str
    metadata: DataRequestMeta | None


class Object(TypedDict):
    code: str
    mime_type: str
    size_bytes: int
    name: str
    event: ItemCode | None
    project: ItemCode
    access_url: str | None
    access_expires: str


class EventMetric(TypedDict):
    channel: Channel
    value: float | str | dict | list | None


class EventTag(TypedDict):
    tag: str
    value: Any | None


class Event(TypedDict):
    code: str
    project: Project
    start_time: str
    end_time: str
    objects: list[Object]
    metrics: list[EventMetric]
    tags: list[EventTag]


UserContactMethod = Literal['EMAIL', 'SMS']


class ContactGroup(TypedDict):
    project: str
    code: str
    label: str
    users: dict[str, list[UserContactMethod]]


class ScheduledReport(TypedDict):
    code: str
    project: str
    label: str
    revision: str
    schedule: Optional[str]
    contact_group: Optional[str]
    last_scheduled: Optional[str]


class ScheduledReportLog(TypedDict):
    code: str
    report: str
    scheduled_start: Optional[str]
    actual_start: str
    actual_finish: Optional[str]
    status: Literal['IN_PROGRESS', 'COMPLETED', 'FAILED']
    message: Optional[str]
    access_url: Optional[str]
    mime_type: Optional[str]
    filename: Optional[str]


class DatatableColumn(TypedDict):
    code: str
    column_label: str


class DatatableOut(TypedDict):
    code: str
    name: str
    src: Optional[str]
    enabled: bool
    sampling_period: Optional[str]
    columns: list[DatatableColumn]


class Datalogger(TypedDict):
    code: str
    project: ItemCode
    name: str
    type: str
    uri: str
    datatables: list[DatatableOut]


class RainflowSource(TypedDict):
    linked_primary_channel: Channel
    sub_channels: list[Channel]


class RainflowConfiguration(TypedDict):
    project: str
    max_bins: int
    bin_size: float
    multiplier: float
    sources: list[RainflowSource]


class FatigueConnection(TypedDict):
    code: str
    label: str
    multiplier: float
    c_d: float
    m: float
    s_0: float

    bs7608_failure_probability: Optional[float]
    bs7608_detail_category: Optional[str]

    initial_date: str
    initial_damage: float
    sources: list[Channel]

    sub_channels: list[Channel]


class Condition(TypedDict):
    code: str
    source: Channel
    description: str
    upper_exclusive_bound: Optional[float]
    lower_inclusive_bound: Optional[float]
    neutral_position: float


class AlertConfiguration(TypedDict):
    code: str
    project: str
    label: str
    conditions: list[Condition]
    contact_group: Optional[ContactGroup]
    retrigger_interval: str


class AlertLogConditionEntry(TypedDict):
    condition: Condition
    start_value: float
    start_time: str
    start_percentile: float

    peak_value: float
    peak_time: str
    peak_percentile: float

    end_value: float
    end_time: str
    end_percentile: float


class AlertLogComment(TypedDict):
    user_code: str
    comment: str
    created_at: str


class AlertLog(TypedDict):
    code: str
    project: str
    event: Optional[str]
    acknowledged: bool
    fired_at: str
    configuration: str
    conditions: list[AlertLogConditionEntry]
    comments: list[AlertLogComment]


class ListAlertsResponseType(TypedDict):
    alerts: list[AlertLog]
    total: int


class Camera(TypedDict):
    code: str
    project: str
    label: str


class Video(TypedDict):
    code: str
    project: str
    camera: str | None
    start_time: str
    end_time: str
    mime_type: str
    size_bytes: int
    name: str
    event: str | None
    access_url: str | None
    access_expires: str


class Image(TypedDict):
    code: str
    project: str
    camera: str | None
    timestamp: str | None
    mime_type: str
    size_bytes: int
    name: str
    event: str | None
    access_url: str | None
    access_expires: str



import argparse
import fnmatch
import itertools
import logging
import logging.handlers
import os
import sys
import time
from typing import Callable, TypeVar

import schedule

from .. import MercutoClient, MercutoHTTPException
from ..types import DataSample
from .ftp import simple_ftp_server
from .parsers import detect_parser
from .processor import FileProcessor
from .util import batched, get_free_space_excluding_files, get_my_public_ip

logger = logging.getLogger(__name__)

NON_RETRYABLE_ERRORS = {400, 404, 409}  # HTTP status codes that indicate non-retryable errors


class MercutoIngester:
    def __init__(self, project_code: str, api_key: str, hostname: str = 'https://api.rockfieldcloud.com.au') -> None:
        self._client = MercutoClient(url=hostname)
        self._api_key = api_key
        with self._client.as_credentials(api_key=api_key) as client:
            self._project = client.projects().get_project(project_code)
            assert self._project['code'] == project_code

            self._secondary_channels = client.channels().get_channels(project_code, classification='SECONDARY')
            self._datatables = list(itertools.chain.from_iterable([dt['datatables'] for dt in client.devices().list_dataloggers(project_code)]))

        self._channel_map = {c['label']: c['code'] for c in self._secondary_channels}

    def update_mapping(self, mapping: dict[str, str]) -> None:
        """
        Update the channel label to channel code mapping.
        """
        self._channel_map.update(mapping)
        logger.info(f"Updated channel mapping: {self._channel_map}")

    @property
    def project_code(self) -> str:
        return self._project['code']

    def ping(self) -> None:
        """
        Ping the Mercuto serverto update the last seen IP address.
        """
        ip = get_my_public_ip()
        with self._client.as_credentials(api_key=self._api_key) as client:
            client.projects().ping_project(self.project_code, ip_address=ip)
            logging.info(f"Pinged Mercuto server from IP: {ip} for project: {self.project_code}")

    def matching_datatable(self, filename: str) -> str | None:
        """
        Check if any datatables on the project match this file name.
        Returns the datatable code if a match is found, otherwise None.
        """
        basename = os.path.basename(filename)

        def matches(test: str) -> bool:
            """
            test should be a pattern or a filename.
            E.g. "my_data.csv" or "my_data*.csv", or "/path/to/my_data*.csv"
            Do wildcard matching as well as prefix matching.
            """
            test_base = os.path.basename(test)
            if fnmatch.fnmatch(basename, test_base):
                return True
            lhs, _ = os.path.splitext(test_base)
            if basename.startswith(lhs):
                return True
            return False

        for dt in self._datatables:
            # Match using datatable pattern
            if matches(dt['name']):
                return dt['code']
            if dt['src'] and matches(dt['src']):
                return dt['code']
        return None

    def _upload_samples(self, samples: list[DataSample]) -> bool:
        """
        Upload samples to the Mercuto project.
        """
        try:
            with self._client.as_credentials(api_key=self._api_key) as client:
                for batch in batched(samples, 500):
                    client.data().upload_samples(batch)
            return True
        except MercutoHTTPException as e:
            if e.status_code in NON_RETRYABLE_ERRORS:
                logger.exception(
                    "Error indicates bad file that should not be retried. Skipping.")
                return True
            else:
                return False

    def _upload_file(self, file_path: str, datatable_code: str) -> bool:
        """
        Upload a file to the Mercuto project.
        """
        logging.info(f"Uploadeding file {file_path} to datatable {datatable_code} in project {self.project_code}")
        try:
            with self._client.as_credentials(api_key=self._api_key) as client:
                client.data().upload_file(
                    project=self.project_code,
                    datatable=datatable_code,
                    file=file_path,
                )
            return True
        except MercutoHTTPException as e:
            if e.status_code in NON_RETRYABLE_ERRORS:
                logger.exception(
                    "Error indicates bad file that should not be retried. Skipping.")
                return True
            else:
                return False

    def process_file(self, file_path: str) -> bool:
        """
        Process the received file.
        """
        logging.info(f"Processing file: {file_path}")
        datatable_code = self.matching_datatable(file_path)
        if datatable_code:
            logger.info(f"Matched datatable code: {datatable_code} for file: {file_path}")
            return self._upload_file(file_path, datatable_code)
        else:
            parser = detect_parser(file_path)
            samples = parser(file_path, self._channel_map)
            if not samples:
                logging.warning(f"No samples found in file: {file_path}")
                return True
            return self._upload_samples(samples)


T = TypeVar('T')


def call_and_log_error(func: Callable[[], T]) -> T | None:
    """
    Call a function and log any exceptions that occur.
    """
    try:
        return func()
    except Exception:
        logging.exception(f"Error in {func.__name__}")
        return None


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Mercuto Ingester CLI')
    parser.add_argument('-p', '--project', type=str,
                        required=True, help='Mercuto project code')
    parser.add_argument('-k', '--api-key', type=str,
                        required=True, help='API key for Mercuto')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Enable verbose output')
    parser.add_argument('-d', '--directory', type=str,
                        help='Directory to store ingested files. Default is a directory called `buffered-files` in the workdir.')
    parser.add_argument('-s', '--size', type=int,
                        help='Size in MB for total amount of files to store in the buffer. \
                            Default is 75% of the available disk space on the buffer partition excluding the directory itself', default=None)
    parser.add_argument('--max-attempts', type=int,
                        help='Maximum number of attempts to process a file before giving up. Default is 1000.',
                        default=1000)
    parser.add_argument('--workdir', type=str,
                        help='Working directory for the ingester. Default is ~/.mercuto-ingester',)
    parser.add_argument('--logfile', type=str,
                        help='Log file path. No logs written if not provided. Maximum of 4 log files of 1MB each will be kept.\
                            Default is log.txt in the workdir.')
    parser.add_argument('--mapping', type=str,
                        help='Path to a JSON file with channel label to channel code mapping.\
                            If not provided, the ingester will try to detect the channels from the project.',
                        default=None)
    parser.add_argument('--hostname', type=str,
                        help='Hostname to use for the Mercuto server. Default is "https://api.rockfieldcloud.com.au".',
                        default='https://api.rockfieldcloud.com.au')
    parser.add_argument('--clean',
                        help='Drop the database before starting. This will not remove any buffer files and will rescan them on startup.',
                        action='store_true')
    parser.add_argument('--username', type=str,
                        help='Username for the FTP server. Default is "logger".',
                        default='logger')
    parser.add_argument('--password', type=str,
                        help='Password for the FTP server. Default is "password".',
                        default='password')
    parser.add_argument('--port', type=int,
                        help='Port for the FTP server. Default is 2121.',
                        default=2121)
    parser.add_argument('--no-rename', action='store_true',
                        help='Add the current timestamp to the end of the files received via FTP. \
                        This is useful to avoid overwriting files with the same name.')

    args = parser.parse_args()

    if args.workdir is None:
        workdir = os.path.join(os.path.expanduser('~'), ".mercuto-ingester")
    else:
        workdir = args.workdir
        if not os.path.exists(args.workdir):
            raise ValueError(f"Work directory {args.workdir} does not exist")
    os.makedirs(workdir, exist_ok=True)

    if args.verbose:
        level = logging.DEBUG
    else:
        level = logging.INFO

    handlers: list[logging.Handler] = []
    handlers.append(logging.StreamHandler(sys.stderr))

    if args.logfile is not None:
        logfile = args.logfile
    else:
        logfile = os.path.join(workdir, 'log.txt')
    handlers.append(logging.handlers.RotatingFileHandler(
        logfile, maxBytes=1000000, backupCount=3))

    logging.basicConfig(format='[PID %(process)d] %(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                        datefmt='%d/%m/%Y %H:%M:%S',
                        level=level,
                        handlers=handlers)

    if args.directory is None:
        buffer_directory = os.path.join(workdir, "buffered-files")
    else:
        buffer_directory = args.directory
    os.makedirs(buffer_directory, exist_ok=True)

    ftp_dir = os.path.join(workdir, 'temp-ftp-data')
    os.makedirs(ftp_dir, exist_ok=True)

    size = args.size
    if size is None:
        size = get_free_space_excluding_files(buffer_directory) * 0.75 // (1024 * 1024)  # Convert to MB
        logging.info(f"Buffer size set to {size} MB based on available disk space.")

    if args.mapping is not None:
        import json
        with open(args.mapping, 'r') as f:
            mapping = json.load(f)
        if not isinstance(mapping, dict):
            raise ValueError(f"Mapping file {args.mapping} must contain a JSON object")
    else:
        mapping = {}

    logger.info(f"Using work directory: {workdir}")

    database_path = os.path.join(workdir, "buffer.db")
    if args.clean and os.path.exists(database_path):
        logging.info(f"Dropping existing database at {database_path}")
        os.remove(database_path)

    ingester = MercutoIngester(
        project_code=args.project,
        api_key=args.api_key,
        hostname=args.hostname)

    ingester.update_mapping(mapping)

    processor = FileProcessor(
        buffer_dir=buffer_directory,
        db_path=database_path,
        process_callback=ingester.process_file,
        max_attempts=args.max_attempts,
        free_space_mb=size)

    processor.scan_existing_files()

    with simple_ftp_server(directory=buffer_directory,
                           username=args.username, password=args.password, port=args.port,
                           callback=processor.add_file_to_db, rename=not args.no_rename,
                           workdir=workdir):
        schedule.every(60).seconds.do(call_and_log_error, ingester.ping)
        schedule.every(5).seconds.do(call_and_log_error, processor.process_next_file)
        schedule.every(2).minutes.do(call_and_log_error, processor.cleanup_old_files)

        while True:
            schedule.run_pending()
            time.sleep(0.5)



import base64
import contextlib
import logging
import mimetypes
import os
import time
import urllib.parse
from contextlib import nullcontext
from datetime import datetime, timedelta
from typing import (Any, BinaryIO, Iterable, Iterator, List, Literal, Mapping,
                    Optional, Sequence, TextIO, Type, TypeVar)

import requests
import requests.cookies

from ._util import timedelta_isoformat
from .exceptions import MercutoClientException, MercutoHTTPException
from .types import (CHANNEL_CLASSIFICATION, AlertConfiguration,
                    AuthHealthcheckResult, Camera, Channel, Condition,
                    ContactGroup, Dashboards, Datalogger, DataRequest,
                    DataSample, Device, DeviceType, Event, FatigueConnection,
                    Image, ListAlertsResponseType, NewUserApiKey, Object,
                    PermissionGroup, Project, RainflowConfiguration,
                    ScheduledReport, ScheduledReportLog,
                    SystemHealthcheckResult, Tenant, Units, User,
                    UserContactMethod, UserDetails, VerifyMeResult, Video)

logger = logging.getLogger(__name__)


class IAuthenticationMethod:
    def update_header(self, header: dict[str, str]) -> None:
        return

    def unique_key(self) -> str:
        raise NotImplementedError(
            f"unique_key not implemented for type {self.__class__.__name__}")


class ApiKeyAuthentication(IAuthenticationMethod):
    def __init__(self, api_key: str) -> None:
        self.api_key = api_key

    def update_header(self, header: dict[str, str]) -> None:
        header['X-Api-Key'] = self.api_key

    def unique_key(self) -> str:
        return f'api-key:{self.api_key}'


class ServiceTokenAuthentication(IAuthenticationMethod):
    def __init__(self, service_token: str) -> None:
        self.service_token = service_token

    def update_header(self, header: dict[str, str]) -> None:
        header['X-Service-Token'] = self.service_token

    def unique_key(self) -> str:
        return f'service-token:{self.service_token}'


class AuthorizationHeaderAuthentication(IAuthenticationMethod):
    def __init__(self, authorization_header: str) -> None:
        self.authorization_header = authorization_header

    def update_header(self, header: dict[str, str]) -> None:
        header['Authorization'] = self.authorization_header

    def unique_key(self) -> str:
        return f'auth-bearer:{self.authorization_header}'


class NullAuthenticationMethod(IAuthenticationMethod):
    def update_header(self, header: dict[str, str]) -> None:
        pass

    def unique_key(self) -> str:
        return 'null-authentication'


def create_authentication_method(api_key: Optional[str] = None,
                                 service_token: Optional[str] = None,
                                 headers: Optional[Mapping[str, str]] = None) -> IAuthenticationMethod:
    if api_key is not None and service_token is not None and headers is not None:
        raise MercutoClientException(
            "Only one of api_key or service_token can be provided")
    authorization_header = None
    if headers is not None:
        api_key = headers.get('X-Api-Key', None)
        service_token = headers.get('X-Service-Token', None)
        authorization_header = headers.get('Authorization', None)
    if api_key is not None:
        return ApiKeyAuthentication(api_key)
    elif service_token is not None:
        return ServiceTokenAuthentication(service_token)
    elif authorization_header is not None:
        return AuthorizationHeaderAuthentication(authorization_header)
    else:
        return NullAuthenticationMethod()


class Module:
    def __init__(self, client: 'MercutoClient') -> None:
        self._client = client

    @property
    def client(self) -> 'MercutoClient':
        return self._client


T = TypeVar('T', bound='Module')


class MercutoClient:
    def __init__(self, url: Optional[str] = None, verify_ssl: bool = True, active_session: Optional[requests.Session] = None) -> None:
        if url is None:
            url = os.environ.get(
                'MERCUTO_API_URL', 'https://api.rockfieldcloud.com.au')
        assert isinstance(url, str)

        if url.endswith('/'):
            url = url[:-1]

        if not url.startswith('https://'):
            raise ValueError(f'Url must be https, is {url}')

        self._url = url
        self._verify_ssl = verify_ssl

        if active_session is None:
            self._current_session = requests.Session()
        else:
            self._current_session = active_session

        self._auth_method: Optional[IAuthenticationMethod] = None
        self._cookies = requests.cookies.RequestsCookieJar()

        self._modules: dict[str, Module] = {}

    def url(self) -> str:
        return self._url

    def credentials_key(self) -> str:
        """
        Generate a unique key that identifies the current credentials set.
        """
        if self._auth_method is None:
            raise MercutoClientException("No credentials set")
        return self._auth_method.unique_key()

    def set_verify_ssl(self, verify_ssl: bool) -> None:
        self._verify_ssl = verify_ssl

    def copy(self) -> 'MercutoClient':
        return MercutoClient(self._url, self._verify_ssl, self._current_session)

    @contextlib.contextmanager
    def as_credentials(self, api_key: Optional[str] = None,
                       service_token: Optional[str] = None,
                       headers: Optional[Mapping[str, str]] = None) -> Iterator['MercutoClient']:
        """
        Same as .connect(), but as a context manager. Will automatically logout when exiting the context.
        """
        # TODO: We are passing the current session along to re-use connections for speed. Will this cause security issues?
        other = MercutoClient(self._url, self._verify_ssl,
                              self._current_session)
        try:
            yield other.connect(api_key=api_key, service_token=service_token, headers=headers)
        finally:
            other.logout()

    def connect(self, *, api_key: Optional[str] = None,
                service_token: Optional[str] = None,
                headers: Optional[Mapping[str, str]] = None) -> 'MercutoClient':
        """
        Attempt to connect using any available method.
        if api_key is provided, use the api_key.
        if service_token is provided, use the service_token.
        if headers is provided, attempt to extract either api_key or service_token from given header set.
            headers should be a dictionary of headers that would be sent in a request. Useful for using existing authenation mechanism for forwarding.

        """
        authentication = create_authentication_method(
            api_key=api_key, service_token=service_token, headers=headers)
        self.login(authentication)
        return self

    def _update_headers(self, headers: dict[str, str]) -> dict[str, str]:
        base: dict[str, str] = {}

        if self._auth_method is not None:
            self._auth_method.update_header(base)
        base.update(headers)
        return base

    def build_url(self, path: str, **params: Any) -> str:
        if path.startswith('/'):
            path = path[1:]
        if path.endswith('/'):
            path = path[:-1]
        return f"{self._url}/{path}/?{urllib.parse.urlencode(params)}"

    def _request_json(self, method: str, url: str, *args: Any, **kwargs: Any) -> Any:
        if 'timeout' not in kwargs:
            kwargs['timeout'] = 10
        kwargs['headers'] = self._update_headers(kwargs.get('headers', {}))

        if 'verify' not in kwargs:
            kwargs['verify'] = self._verify_ssl

        if 'cookies' not in kwargs:
            kwargs['cookies'] = self._cookies
        return self._make_request(method, url, *args, **kwargs)

    def _make_request(self, method: str, url: str, *args: Any, **kwargs: Any) -> Any:
        if url.startswith('/'):
            url = url[1:]
        full_url = f"{self._url}/{url}"
        start = time.time()
        resp = self._current_session.request(method, full_url, *args, **kwargs)
        duration = time.time() - start
        logger.debug("Made request to %s %s in %.2f seconds (code=%s)",
                     method, full_url, duration, resp.status_code)
        if not resp.ok:
            raise MercutoHTTPException(resp.text, resp.status_code)
        if resp.headers.get('Content-Type', '') != 'application/json':
            raise MercutoClientException(f"Response is not JSON: {resp.text}")
        resp.cookies.update(self._cookies)
        return resp.json()

    def request(self, method: str, url: str, *args: Any, **kwargs: Any) -> Any:
        return self._request_json(method, url, *args, **kwargs)

    def add_and_fetch_module(self, name: str, module: Type[T]) -> T:
        if name not in self._modules:
            self._modules[name] = module(self)
        return self._modules[name]  # type: ignore

    def identity(self) -> 'MercutoIdentityService':
        return self.add_and_fetch_module('identity', MercutoIdentityService)

    def projects(self) -> 'MercutoProjectService':
        return self.add_and_fetch_module('projects', MercutoProjectService)

    def fatigue(self) -> 'MercutoFatigueService':
        return self.add_and_fetch_module('fatigue', MercutoFatigueService)

    def channels(self) -> 'MercutoChannelService':
        return self.add_and_fetch_module('channels', MercutoChannelService)

    def data(self) -> 'MercutoDataService':
        return self.add_and_fetch_module('data', MercutoDataService)

    def events(self) -> 'MercutoEventService':
        return self.add_and_fetch_module('events', MercutoEventService)

    def alerts(self) -> 'MercutoAlertService':
        return self.add_and_fetch_module('alerts', MercutoAlertService)

    def devices(self) -> 'MercutoDeviceService':
        return self.add_and_fetch_module('devices', MercutoDeviceService)

    def notifications(self) -> 'MercutoNotificationsService':
        return self.add_and_fetch_module('notifications', MercutoNotificationsService)

    def reports(self) -> 'MercutoReportingService':
        return self.add_and_fetch_module('reports', MercutoReportingService)

    def objects(self) -> 'MercutoObjectService':
        return self.add_and_fetch_module('objects', MercutoObjectService)

    def media(self) -> 'MercutoMediaService':
        return self.add_and_fetch_module('media', MercutoMediaService)

    def login(self, authentication: IAuthenticationMethod) -> None:
        self._auth_method = authentication

    def logout(self) -> None:
        self._auth_method = None

    def healthcheck(self) -> SystemHealthcheckResult:
        return self._request_json('GET', '/healthcheck')  # type: ignore[no-any-return]


class MercutoDataService(Module):
    def __init__(self, client: 'MercutoClient') -> None:
        super().__init__(client)

    def upload_samples(self, samples: Sequence[DataSample]) -> None:
        self._client._request_json(
            'POST', '/data/upload/samples', json=samples)

    def upload_file(self, project: str, datatable: str, file: str | bytes | TextIO | BinaryIO, filename: Optional[str] = None) -> None:
        if isinstance(file, str):
            ctx = open(file, 'rb')
            filename = filename or os.path.basename(file)
        else:
            ctx = nullcontext(file)  # type: ignore
            filename = filename or 'file.dat'

        with ctx as f:
            self._client._request_json('POST', '/files/upload/small', params={'project_code': project, 'datatable_code': datatable},
                                       files={'file': (filename, f, 'text/csv')})

    def get_data_url(
        self,
        project_code: str,
        start_time: datetime | str | None = None,
        end_time: datetime | str | None = None,
        event_code: str | None = None,
        channel_codes: Iterable[str] | None = None,
        primary_channels: bool | None = None,
        channels_like: str | None = None,
        file_format: Literal['DAT', 'CSV', 'PARQUET', 'FEATHER'] = 'DAT',
        frame_format: Literal['RECORDS', 'COLUMNS'] = 'COLUMNS',
        channel_format: Literal['CODE', 'LABEL'] = 'LABEL',
        timeout: timedelta | None = timedelta(seconds=20),
    ) -> str:
        request = self.get_data_request(
            project_code=project_code,
            start_time=start_time,
            end_time=end_time,
            event_code=event_code,
            channel_codes=channel_codes,
            primary_channels=primary_channels,
            channels_like=channels_like,
            file_format=file_format,
            frame_format=frame_format,
            channel_format=channel_format,
            timeout=timeout,
        )
        if request['presigned_url'] is None:
            raise MercutoClientException("No presigned URL available")

        return request['presigned_url']

    def get_data_request(
        self,
        project_code: str,
        start_time: datetime | str | None = None,
        end_time: datetime | str | None = None,
        event_code: str | None = None,
        channel_codes: Iterable[str] | None = None,
        primary_channels: bool | None = None,
        channels_like: str | None = None,
        file_format: Literal['DAT', 'CSV', 'PARQUET', 'FEATHER'] = 'DAT',
        frame_format: Literal['RECORDS', 'COLUMNS'] = 'COLUMNS',
        channel_format: Literal['CODE', 'LABEL'] = 'LABEL',
        timeout: timedelta | None = timedelta(seconds=20),
    ) -> DataRequest:
        params = {
            'timeout': 0
        }

        body = {
            'project_code': project_code,
            'start_time': start_time.isoformat() if isinstance(start_time, datetime) else start_time,
            'end_time': end_time.isoformat() if isinstance(end_time, datetime) else end_time,
            'event_code': event_code,
            'channel_codes': channel_codes,
            'primary_channels': primary_channels,
            'channels_like': channels_like,
            'data_format': file_format,
            'frame_format': frame_format,
            'channel_format': channel_format,
        }

        request: DataRequest = self._client._request_json(
            'POST',
            '/data/requests',
            params=params,
            json=body)

        start = time.time()
        while True:
            if timeout is not None and time.time() - start > timeout.total_seconds():
                raise MercutoClientException(
                    "Timeout waiting for data request")
            if request['completed_at'] is not None:
                return request

            time.sleep(1)
            request = self._client._request_json(
                'GET', f'/data/requests/{request["code"]}')


class MercutoEventService(Module):
    def __init__(self, client: 'MercutoClient') -> None:
        super().__init__(client)

    def list_events(self, project: str) -> list[Event]:
        params: dict[str, str] = {'project_code': project}
        return self._client._request_json('GET', '/events', params=params)

    def get_nearest_event(
        self,
        project_code: str,
        to: datetime | str,
        maximum_delta: timedelta | None = None,
    ) -> Event:
        params = {
            'project_code': project_code,
            'to': to.isoformat() if isinstance(to, datetime) else to,
            'maximum_delta': timedelta_isoformat(maximum_delta) if maximum_delta is not None else None,
        }

        return self.client._request_json('GET', '/events/nearest', params=params)

    def get_nearest_event_url(
            self,
            project_code: str,
            to: datetime | str,
            maximum_delta: timedelta | None = None,
            file_format: Literal['DAT', 'CSV', 'PARQUET', 'FEATHER'] = 'DAT',
            frame_format: Literal['COLUMNS', 'RECORDS'] = 'COLUMNS',
    ) -> str:
        event = self.get_nearest_event(project_code, to, maximum_delta)

        return self.client.data().get_data_url(
            project_code=project_code,
            event_code=event['code'],
            primary_channels=True,
            file_format=file_format,
            frame_format=frame_format)


class MercutoAlertService(Module):
    def __init__(self, client: 'MercutoClient') -> None:
        super().__init__(client)

    def get_condition(self, code: str) -> Condition:
        return self._client._request_json('GET', f'/alerts/conditions/{code}')

    def create_condition(self, source: str, description: str, *,
                         lower_bound: Optional[float] = None,
                         upper_bound: Optional[float] = None,
                         neutral_position: float = 0) -> Condition:
        json = {
            'source_channel_code': source,
            'description': description,
            'neutral_position': neutral_position
        }
        if lower_bound is not None:
            json['lower_inclusive_bound'] = lower_bound
        if upper_bound is not None:
            json['upper_exclusive_bound'] = upper_bound
        return self._client._request_json('PUT', '/alerts/conditions', json=json)

    def create_configuration(self, label: str, conditions: list[str], contact_group: Optional[str] = None) -> AlertConfiguration:
        json = {
            'label': label,
            'conditions': conditions,

        }
        if contact_group is not None:
            json['contact_group'] = contact_group
        return self._client._request_json('PUT', '/alerts/configurations', json=json)

    def get_alert_configuration(self, code: str) -> AlertConfiguration:
        return self._client._request_json('GET', f'/alerts/configurations/{code}')

    def list_logs(
            self,
            project: str | None = None,
            configuration: str | None = None,
            channels: list[str] | None = None,
            start_time: datetime | str | None = None,
            end_time: datetime | str | None = None,
            limit: int = 10,
            offset: int = 0,
            latest_only: bool = False,
    ) -> ListAlertsResponseType:
        params = {
            'project': project,
            'configuration_code': configuration,
            'channels': channels,
            'start_time': start_time.isoformat() if isinstance(start_time, datetime) else start_time,
            'end_time': end_time.isoformat() if isinstance(end_time, datetime) else end_time,
            'limit': limit,
            'offset': offset,
            'latest_only': latest_only,
        }

        return self._client._request_json('GET', '/alerts/logs', params=params)


class MercutoProjectService(Module):
    def __init__(self, client: 'MercutoClient') -> None:
        super().__init__(client)

    def get_project(self, code: str) -> Project:
        return self._client._request_json('GET', f'/projects/{code}')

    def get_projects(self) -> list[Project]:
        return self._client._request_json('GET', '/projects')

    def create_project(self, name: str, project_number: str, description: str, tenant: str,
                       timezone: str, latitude: Optional[float] = None,
                       longitude: Optional[float] = None) -> Project:

        return self._client._request_json('PUT', '/projects',
                                          json={'name': name, 'project_number': project_number, 'description': description,
                                                'tenant_code': tenant,
                                                'timezone': timezone,
                                                'latitude': latitude,
                                                'longitude': longitude,
                                                'project_type': 1,
                                                'channels': []})

    def ping_project(self, project: str, ip_address: str) -> None:
        self._client._request_json(
            'POST', f'/projects/{project}/ping', json={'ip_address': ip_address})

    def create_channel(self, project: str, label: str,  sampling_period: timedelta) -> Channel:
        return self._client._request_json('PUT', '/channels', json={
            'project_code': project,
            'label': label,
            'classification': 'SECONDARY',
            'sampling_period': timedelta_isoformat(sampling_period),
        })

    def create_dashboard(self, project_code: str, dashboards: Dashboards) -> bool:
        return self._client._request_json('POST', f'/projects/{project_code}/dashboard', json=dashboards) is None


class MercutoFatigueService(Module):
    def __init__(self, client: 'MercutoClient') -> None:
        super().__init__(client)

    def setup_rainflow(self, project: str,
                       max_bins: int,
                       bin_size: float,
                       multiplier: float,
                       channels: list[str],
                       reservoir_adjustment: bool = True,
                       enable_root_mean_cubed: bool = True,
                       enable_root_sum_cubed: bool = True) -> RainflowConfiguration:
        return self.client._request_json('PUT', '/fatigue/rainflow/setup', json=dict(
            project=project,
            max_bins=max_bins,
            bin_size=bin_size,
            multiplier=multiplier,
            reservoir_adjustment=reservoir_adjustment,
            channels=channels,
            enable_root_mean_cubed=enable_root_mean_cubed,
            enable_root_sum_cubed=enable_root_sum_cubed
        ))

    def add_connection(self, project: str, label: str,
                       multiplier: float, c_d: float, m: float, s_0: float,
                       bs7608_failure_probability: float, bs7608_detail_category: str,
                       initial_date: datetime, initial_damage: float,
                       sources: list[str]) -> FatigueConnection:
        """
        Sources should be a list of Primary Channel codes.
        """
        return self.client._request_json('PUT', '/fatigue/connections', json=dict(
            project=project,
            label=label,
            multiplier=multiplier,
            c_d=c_d,
            m=m,
            s_0=s_0,
            bs7608_failure_probability=bs7608_failure_probability,
            bs7608_detail_category=bs7608_detail_category,
            initial_date=initial_date.isoformat(),
            initial_damage=initial_damage,
            sources=sources
        ))


class MercutoChannelService(Module):
    def __init__(self, client: MercutoClient) -> None:
        super().__init__(client)

    def get_units(self) -> List[Units]:
        return self._client._request_json('GET', '/channels/units')

    def create_units(self, name: str, unit: str) -> Units:
        return self._client._request_json('PUT', '/channels/units', json=dict(name=name, unit=unit))

    def create_channel(self, project_code: str, label: str, classification: CHANNEL_CLASSIFICATION, sampling_period: timedelta) -> Channel:
        return self._client._request_json('PUT', '/channels', json=dict(
            project_code=project_code,
            label=label,
            classification=classification,
            sampling_period=timedelta_isoformat(sampling_period)))

    def modify_channel(self,
                       channel: str,
                       label:  Optional[str] = None,
                       units_code:  Optional[str] = None,
                       device_code:  Optional[str] = None,
                       metric:  Optional[str] = None,
                       multiplier: float = 1,
                       offset: float = 0) -> Channel:
        data: dict = {}
        if label is not None:
            data['label'] = label
        if units_code is not None:
            data['units_code'] = units_code
        if metric is not None:
            data['metric'] = metric
        if device_code is not None:
            data['device_code'] = device_code
        data['multiplier'] = multiplier
        data['offset'] = offset
        return self._client._request_json('PATCH', f'/channels/{channel}', json=data)

    def get_channels(
        self,
        project_code: str | None = None,
        classification: CHANNEL_CLASSIFICATION | None = None,
        aggregate: str | None = None,
        metric: str | None = None,
    ) -> list[Channel]:
        limit = 200
        offset = 0
        channels: list[Channel] = []

        while True:
            params = {
                'project_code': project_code,
                'classification': classification,
                'aggregate': aggregate,
                'metric': metric,
                'limit': limit,
                'offset': offset,
            }

            resp = self._client._request_json(
                'GET', '/channels', params=params)
            channels.extend(resp)

            if len(resp) < limit:
                break

            offset += limit

        return channels


class MercutoDeviceService(Module):
    def __init__(self, client: MercutoClient) -> None:
        super().__init__(client)

    def get_device_types(self) -> list[DeviceType]:
        return self._client._request_json('GET', '/devices/types')

    def create_device_type(self, description: str, manufacturer: str, model_number: str) -> DeviceType:
        return self._client._request_json('PUT', '/devices/types', json=dict(
            description=description,
            manufacturer=manufacturer,
            model_number=model_number))

    def get_devices(self, project_code: str, limit: int, offset: int) -> list[Device]:
        return self._client._request_json('GET', '/devices', params=dict(project_code=project_code, limit=limit, offset=offset))

    def get_device(self, device_code: str) -> Device:
        return self._client._request_json('GET', f'/devices/{device_code}')

    def create_device(self,
                      project_code: str,
                      label: str,
                      device_type_code: str,
                      groups: list[str],
                      location_description: Optional[str] = None) -> Device:
        return self._client._request_json('PUT', '/devices', json=dict(
            project_code=project_code,
            label=label,
            device_type_code=device_type_code,
            groups=groups,
            location_description=location_description))

    def list_dataloggers(self, project: str) -> list[Datalogger]:
        return self._client._request_json('GET', '/dataloggers', params={'project_code': project})


class MercutoIdentityService(Module):
    def __init__(self, client: 'MercutoClient') -> None:
        super().__init__(client)

    def create_tenant(self, name: str, description: str, logo_url: Optional[str] = None) -> Tenant:
        return self._client._request_json('PUT', '/identity/tenants',
                                          json={'name': name, 'description': description, 'logo_url': logo_url})

    def list_tenants(self) -> list[Tenant]:
        return self._client._request_json('GET', '/identity/tenants')

    def get_tenant(self, code: str) -> Tenant:
        return self._client._request_json('GET', f'/identity/tenants/{code}')

    def create_user(self, username: str, tenant: str, description: str,
                    group: str, password: Optional[str] = None) -> User:
        return self._client._request_json('PUT', '/identity/users',
                                          json={'username': username, 'tenant_code': tenant, 'description': description,
                                                'group_code': group, 'default_password': password})

    def list_users(self, project: Optional[str] = None,
                   tenant: Optional[str] = None) -> list[User]:
        params = {}
        if project is not None:
            params['project'] = project
        if tenant is not None:
            params['tenant'] = tenant
        return self._client._request_json('GET', '/identity/users', params=params)

    def get_user(self, code: str) -> User:
        return self._client._request_json('GET', f'/identity/users/{code}')

    def get_user_details(self, code: str) -> UserDetails:
        return self._client._request_json('GET', f'/identity/users/{code}/details')

    def edit_user_details(self, code: str, first_name: Optional[str], last_name: Optional[str],
                          email_address: Optional[str], mobile_number: Optional[str]) -> UserDetails:
        return self._client._request_json('PATCH', f'/identity/users/{code}/details', json={
            'first_name': first_name,
            'last_name': last_name,
            'email_address': email_address,
            'mobile_number': mobile_number
        })

    def generate_api_key(self, user: str, description: str) -> NewUserApiKey:
        return self._client._request_json('POST', f'/identity/users/{user}/api_keys',
                                          json={'description': description})

    def grant_user_permission(self, user: str, resource: str, action: str) -> None:
        return self._client._request_json('POST', f'/identity/users/{user}/grant',
                                          json={'resource': resource, 'action': action})

    def create_permission_group(self, tenant: str, label: str,
                                acl_json: str) -> PermissionGroup:
        return self._client._request_json('PUT', '/identity/permissions', json={
            'tenant': tenant,
            'label': label,
            'acl_policy': acl_json
        })

    def list_permission_groups(self) -> list[PermissionGroup]:
        return self._client._request_json('GET', '/identity/permissions')

    def get_permission_group(self, code: str) -> PermissionGroup:
        return self._client._request_json('GET', f'/identity/permissions/{code}')

    def update_permission_group(self, code: str, label: str,
                                acl_json: str) -> None:
        return self._client._request_json('PATCH', f'/identity/permissions/{code}', json={
            'label': label,
            'acl_policy': acl_json
        })

    def verify_me(self) -> VerifyMeResult:
        return self._client._request_json('GET', '/identity/verify/me')

    def healthcheck(self) -> AuthHealthcheckResult:
        return self._client._request_json('GET', '/identity/healthcheck')


class MercutoNotificationsService(Module):
    def __init__(self, client: 'MercutoClient') -> None:
        super().__init__(client)

    def list_contact_groups(self, project: Optional[str] = None) -> list[ContactGroup]:
        params = {}
        if project is not None:
            params['project'] = project
        return self._client._request_json('GET', '/notifications/contact_groups', params=params)

    def get_contact_group(self, code: str) -> ContactGroup:
        return self._client._request_json('GET', f'/notifications/contact_groups/{code}')

    def create_contact_group(self, project: str, label: str, users: dict[str, list[UserContactMethod]]) -> ContactGroup:
        return self._client._request_json('PUT', '/notifications/contact_groups',
                                          json={
                                              'project': project,
                                              'label': label,
                                              'users': users
                                          })


class MercutoReportingService(Module):
    def __init__(self, client: 'MercutoClient') -> None:
        super().__init__(client)

    def list_reports(self, project: Optional[str] = None) -> list['ScheduledReport']:
        params = {}
        if project is not None:
            params['project'] = project
        return self._client._request_json('GET', '/reports/scheduled', params=params)

    def create_report(self, project: str, label: str, schedule: str, revision: str,
                      api_key: Optional[str] = None, contact_group: Optional[str] = None) -> ScheduledReport:
        return self._client._request_json('PUT', '/reports/scheduled', json={
            'project': project,
            'label': label,
            'schedule': schedule,
            'revision': revision,
            'execution_role_api_key': api_key,
            'contact_group': contact_group
        })

    def generate_report(self, report: str, timestamp: datetime, mark_as_scheduled: bool = False) -> ScheduledReportLog:
        return self._client._request_json('PUT', f'/reports/scheduled/{report}/generate', json={
            'timestamp': timestamp.isoformat(),
            'mark_as_scheduled': mark_as_scheduled
        })

    def list_report_logs(self, report: str, project: Optional[str] = None) -> list[ScheduledReportLog]:
        params: dict[str, str] = {}
        if project is not None:
            params['project'] = project
        return self._client._request_json('GET', f'/reports/scheduled/{report}/logs', params=params)

    def get_report_log(self, report: str, log: str) -> ScheduledReportLog:
        return self._client._request_json('GET', f'/reports/scheduled/{report}/logs/{log}')


class MercutoObjectService(Module):
    def __init__(self, client: 'MercutoClient') -> None:
        super().__init__(client)

    def upload_file(self, project_code: str, file: str, event_code: Optional[str] = None,
                    mime_type: Optional[str] = None) -> Object:
        with open(file, 'rb') as f:
            if mime_type is None:
                mime_type = mimetypes.guess_type(file, strict=False)[0]
                if mime_type is None:
                    raise MercutoClientException(
                        f"Could not determine mime type for {file}")
            base64_data = base64.b64encode(f.read()).decode('utf-8')
            data_url = f'data:{mime_type};base64,{base64_data}'

            return self._client._request_json('POST', '/objects/upload', params={
                'project_code': project_code,
                'event_code': event_code
            }, json={
                'filename': os.path.basename(file),
                'mime_type': mime_type,
                'data_url': data_url
            })


class MercutoMediaService(Module):
    def __init__(self, client: 'MercutoClient') -> None:
        super().__init__(client)

    def list_cameras(self, project: str) -> list[Camera]:
        params = {}
        params['project_code'] = project
        return self._client._request_json('GET', '/media/cameras', params=params)

    def list_videos(self, project: Optional[str] = None, event: Optional[str] = None, camera: Optional[str] = None) -> list[Video]:
        params = {}
        if project is not None:
            params['project'] = project
        if event is not None:
            params['event'] = event
        if camera is not None:
            params['camera'] = camera
        return self._client._request_json('GET', '/media/videos', params=params)

    def list_images(self, project: Optional[str] = None, event: Optional[str] = None, camera: Optional[str] = None) -> list[Image]:
        params = {}
        if project is not None:
            params['project'] = project
        if event is not None:
            params['event'] = event
        if camera is not None:
            params['camera'] = camera
        return self._client._request_json('GET', '/media/images', params=params)

    def get_image(self, code: str) -> Image:
        return self._client._request_json('GET', f'/media/images/{code}')  # type: ignore[no-any-return]

    def upload_image(self, project: str, file: str, event: Optional[str] = None,
                     camera: Optional[str] = None, timestamp: Optional[datetime] = None,
                     filename: Optional[str] = None) -> Image:
        if timestamp is not None and timestamp.tzinfo is None:
            raise MercutoClientException("Timestamp must be timezone aware")

        mimetype, _ = mimetypes.guess_type(file, strict=False)
        if mimetype is None or not mimetype.startswith('image/'):
            raise MercutoClientException(f"File {file} is not an image")

        if os.stat(file).st_size > 5_000_000:
            raise MercutoClientException(f"File {file} is too large")

        if filename is None:
            filename = os.path.basename(file)

        params = {}
        params['project'] = project
        if event is not None:
            params['event'] = event
        if camera is not None:
            params['camera'] = camera
        if timestamp is not None:
            params['timestamp'] = timestamp.isoformat()

        with open(file, 'rb') as f:
            return self._client._request_json('PUT', '/media/images', params=params, files={
                'file': (filename, f, mimetype)
            })



import pytest

from .. import MercutoClient
from ..mocks import mock_client


def test_mock_injection_before_client_creation():
    count = 0

    def on_get_healthcheck(*args, **kwargs):
        nonlocal count
        count += 1
        return 'mocked'
    with mock_client() as mock:
        mock.on('GET', '/healthcheck', on_get_healthcheck)
        client = MercutoClient()
        assert client.healthcheck() == 'mocked'
    assert count == 1


def test_mock_injection_after_client_creation():
    count = 0
    client = MercutoClient()

    def on_get_healthcheck(*args, **kwargs):
        nonlocal count
        count += 1
        return 'mocked'
    with mock_client() as mock:
        mock.on('GET', '/healthcheck', on_get_healthcheck)
        assert client.healthcheck() == 'mocked'
    assert count == 1


def test_mock_releases_after_end_of_context():
    client = MercutoClient()
    with mock_client() as mock:
        key = mock.add_user(user='this is a test')
        client.connect(api_key=key)
        assert client.identity().verify_me()['user'] == 'this is a test'

    with pytest.raises(Exception):
        client.identity().verify_me()


def test_mock_verify_me():
    client = MercutoClient()
    with mock_client() as mock:
        with pytest.raises(Exception):
            client.identity().verify_me()

        client.connect(api_key='bad api key')
        with pytest.raises(Exception):
            client.identity().verify_me()

        key = mock.add_user(user='this is a test user',
                            tenant='test-tenant', permission_group='test-group')
        client.connect(api_key=key)
        assert client.identity().verify_me()['user'] == 'this is a test user'
        assert client.identity().verify_me()['tenant'] == 'test-tenant'
        assert client.identity().verify_me()[
            'permission_group'] == 'test-group'

        mock.delete_user(key)
        with pytest.raises(Exception):
            client.identity().verify_me()


def test_mock_get_user():
    client = MercutoClient()
    with mock_client() as mock:
        client.connect(api_key='bad api key')
        with pytest.raises(Exception):
            client.identity().get_user('12345')

        key = mock.add_user(user='code1', tenant='test-tenant', permission_group='test-group',
                            username='testing@example.com')
        client.connect(api_key=key)

        assert client.identity().get_user('code1')['code'] == 'code1'
        assert client.identity().get_user(
            'code1')['username'] == 'testing@example.com'

        mock.delete_user(key)
        with pytest.raises(Exception):
            client.identity().verify_me()


def test_mock_unsupported_endpoint():
    client = MercutoClient()
    with mock_client():
        with pytest.raises(NotImplementedError):
            client.identity().list_tenants()



"""
Copyright (C) 2025 Rockfield Technologies Australia Pty Ltd

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Mapping, Optional

from .client import MercutoClient
from .exceptions import MercutoClientException, MercutoHTTPException

__all__ = ['MercutoClient', 'MercutoHTTPException', 'MercutoClientException']


def connect(api_key: Optional[str] = None,
            service_token: Optional[str] = None,
            headers: Optional[Mapping[str, str]] = None) -> MercutoClient:
    return MercutoClient().connect(api_key=api_key,
                                   service_token=service_token,
                                   headers=headers)



import contextlib
import dataclasses
import re
import uuid
from typing import Any, Optional, Protocol

from .exceptions import MercutoClientException
from .types import Tenant, User, UserDetails, VerifyMeResult

"""
This module provides a context manager for mocking the Mercuto Client.

Within the mock_client() context, any calls to the Mercuto Client will be intercepted and handled by the MercutoMocker.


Usage:

client = MercutoClient()

with mock_client() as mock:
    # Create a user in the system
    api_key = mock.add_user()

    # Login using the generated user key
    client.connect(api_key = api_key)

    # Call Mercuto Client endpoints as normal
    client.identity().verify_me()

"""


class MakeRequestHookType(Protocol):
    def __call__(self, method: str, url: str, *args, **kwargs) -> Any:
        """
        *args and **kwargs are arguments passed to requests.request(method, url, *args, **kwargs)
        """
        pass


@contextlib.contextmanager
def mock_client():
    from .client import MercutoClient
    original = MercutoClient._make_request
    mocker = MercutoMocker()
    try:
        setattr(MercutoClient, '_make_request', mocker._on_make_request)
        yield mocker
    finally:
        setattr(MercutoClient, '_make_request', original)


class MercutoMocker:
    @dataclasses.dataclass
    class MockedUser:
        code: str
        username: str
        description: str
        tenant: str
        permission_group: str
        mobile_number: Optional[str] = None

    def __init__(self) -> None:
        self._hooks: dict[tuple[str, str], MakeRequestHookType] = {}

        # Users based on Api-Key
        self._known_users: dict[str, MercutoMocker.MockedUser] = {}
        self._known_tenants: dict[str, Tenant] = {}

        self._setup_default_hooks()

    def on(self, method: str, path: str, callback: MakeRequestHookType):
        self._hooks[(method, path)] = callback

    def add_tenant(self, code: Optional[str] = None) -> None:
        if code is None:
            code = str(uuid.uuid4())
        self._known_tenants[code] = Tenant(
            code=code,
            name=f"Tenant {code}",
            description=f"Tenant {code}",
            logo_url=None,
        )

    def add_user(self,
                 user: Optional[str] = None,
                 tenant: Optional[str] = None,
                 permission_group: Optional[str] = None,
                 key: Optional[str] = None,
                 username: Optional[str] = None,
                 description: Optional[str] = None,
                 mobile_number: Optional[str] = None,
                 ):
        if user is None:
            # User code
            user = str(uuid.uuid4())
        if tenant is None:
            tenant = str(uuid.uuid4())
        if permission_group is None:
            permission_group = str(uuid.uuid4())
        if key is None:
            key = str(uuid.uuid4())
        if username is None:
            username = f"{user.replace(' ', '')}@example.com"
        if description is None:
            description = f"User {user}"

        mocked = self.MockedUser(user, username, description, tenant, permission_group,
                                 mobile_number=mobile_number)
        self._known_users[key] = mocked
        return key

    def delete_user(self, key: Optional[str] = None, code: Optional[str] = None):
        if key is not None and key in self._known_users:
            del self._known_users[key]

        if code is not None:
            key = next((k for k, u in self._known_users.items()
                       if u.code == code), None)
            if key is not None:
                del self._known_users[key]

    def _mocked_get_tenant(self, method: str, url: str, *args, **kwargs) -> Tenant:
        key = kwargs.get('headers', {}).get('X-Api-Key', None)
        if key is None:
            raise MercutoClientException("No X-Api-Key header provided")
        tenant = self._known_tenants.get(key, None)
        if tenant is None:
            raise MercutoClientException("Tenant not found")
        return tenant

    def _mocked_verify_me(self, method: str, url: str, *args, **kwargs) -> VerifyMeResult:
        key = kwargs.get('headers', {}).get('X-Api-Key', None)
        if key is None:
            raise MercutoClientException("No X-Api-Key header provided")
        user = self._known_users.get(key, None)
        if user is None:
            raise MercutoClientException(f"User {user} not found")
        return VerifyMeResult(
            user=user.code,
            tenant=user.tenant,
            permission_group=user.permission_group,
            acl_policy='{"version": 1, "permissions": []}'
        )

    def _mocked_get_user(self, method: str, url: str, *args, **kwargs) -> User:
        apikey = kwargs.get('headers', {}).get('X-Api-Key', None)
        servicekey = kwargs.get('headers', {}).get('X-Service-Token', None)
        if apikey is None and servicekey is None:
            raise MercutoClientException(
                "No X-Api-Key or X-Service-Token header provided")
        user_code = url.split('/')[-1]
        user = next((u for u in self._known_users.values()
                    if u.code == user_code), None)
        if user is None:
            raise MercutoClientException("User not found")
        return User(
            code=user.code,
            username=user.username,
            description=user.description,
            tenant=user.tenant,
            permission_group=user.permission_group,
        )

    def _mocked_get_user_details(self, method: str, url: str, *args, **kwargs) -> UserDetails:
        apikey = kwargs.get('headers', {}).get('X-Api-Key', None)
        servicekey = kwargs.get('headers', {}).get('X-Service-Token', None)
        if apikey is None and servicekey is None:
            raise MercutoClientException(
                "No X-Api-Key or X-Service-Token header provided")
        url_parts = url.split('/')
        assert url_parts[-1] == 'details'
        user_code = url_parts[-2]
        user = next((u for u in self._known_users.values()
                    if u.code == user_code), None)
        if user is None:
            raise MercutoClientException("User not found")
        return UserDetails(
            code=user.code,
            username=user.username,
            mobile_number=user.mobile_number,
            email_address=None,
            first_name=None,
            last_name=None,
            api_keys=[]
        )

    def _setup_default_hooks(self) -> None:
        self.on('GET', r'\/identity\/verify\/me', self._mocked_verify_me)
        self.on('GET', r'\/identity\/users\/[^\/]+', self._mocked_get_user)
        self.on('GET', r'\/identity\/tenant\/[^\/]+', self._mocked_get_tenant)
        self.on(
            'GET', r'\/identity\/users\/[^\/]+\/details', self._mocked_get_user_details)

    def _on_make_request(self, method: str, url: str, *args, **kwargs) -> Any:

        # First check any custom hooks
        for (hook_method, pattern), callback in self._hooks.items():
            if method == hook_method and re.fullmatch(pattern, url) is not None:
                return callback(method, url, *args, **kwargs)

        raise NotImplementedError(
            "Mocking is not supported for this endpoint: %s %s" % (method, url))



import json


class MercutoClientException(Exception):
    pass


class MercutoHTTPException(MercutoClientException):
    def __init__(self, message: str, status_code: int) -> None:
        super().__init__(message)
        self.status_code = status_code
        self.message = message

    def json(self) -> dict:
        return json.loads(self.message)
